@@ -6,33 +6,32 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
-use std::{
-    fs::File,
-    io::BufReader,
-    os::fd::{AsFd, AsRawFd},
-};
+use std::{fs::File, io::BufReader, os::fd::AsFd};
 
 use libseccomp::ScmpNotifResp;
 use nix::{errno::Errno, fcntl::AtFlags, NixPath};
 
 use crate::{
-    compat::{fstatat64, statx, STATX_BASIC_STATS, STATX_MODE, STATX_TYPE},
-    config::{MAGIC_PREFIX, MMAP_MIN_ADDR},
-    confine::{scmp_arch_bits, EOWNERDEAD},
-    fs::parse_fd,
+    compat::{
+        fstatat64, statx, FileStat64, FileStatx, FileStatxTimestamp, STATX_BASIC_STATS, STATX_MODE,
+        STATX_TYPE,
+    },
+    config::{API_VERSION, MAGIC_PREFIX},
+    confine::{is_valid_ptr, scmp_arch_bits, EOWNERDEAD},
+    fd::parse_fd,
     hash::SydHashSet,
     kernel::{sandbox_path, to_atflags},
     lookup::{CanonicalPath, FileInfo, FileType, FsFlags},
     path::XPath,
     req::{SysArg, SysFlags, UNotifyEventRequest},
-    sandbox::Capability,
+    sandbox::{Capability, Flags},
 };
 
 pub(crate) fn sys_stat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1], req.data.arch) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -52,7 +51,7 @@ pub(crate) fn sys_stat64(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1], req.data.arch) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -70,7 +69,7 @@ pub(crate) fn sys_fstat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1], req.data.arch) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -88,7 +87,7 @@ pub(crate) fn sys_fstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1], req.data.arch) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -104,7 +103,7 @@ pub(crate) fn sys_lstat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1], req.data.arch) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -124,7 +123,7 @@ pub(crate) fn sys_lstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[1] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[1], req.data.arch) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -158,7 +157,7 @@ pub(crate) fn sys_statx(request: UNotifyEventRequest) -> ScmpNotifResp {
     };
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[4] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[4], req.data.arch) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -188,7 +187,7 @@ pub(crate) fn sys_newfstatat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Return EFAULT here for invalid pointers.
-    if req.data.args[2] < *MMAP_MIN_ADDR {
+    if !is_valid_ptr(req.data.args[2], req.data.arch) {
         return request.fail_syscall(Errno::EFAULT);
     }
 
@@ -249,6 +248,8 @@ fn syscall_stat_handler(
         let restrict_stat_bdev = !sandbox.flags.allow_unsafe_stat_bdev();
         let restrict_stat_cdev = !sandbox.flags.allow_unsafe_stat_cdev();
         let mut ghost = false;
+        let caps = *sandbox.state;
+        let sflags = *sandbox.flags;
         if magic {
             if sandbox.locked_for(req.pid()) {
                 // Sandbox is locked, access denied.
@@ -324,7 +325,7 @@ fn syscall_stat_handler(
                 // SAFETY: SysArg.path is None asserting dirfd is Some fd!=AT_FDCWD.
                 let fd = path.dir.as_ref().unwrap();
                 // SAFETY: has_crypt asserts crypt_map is Some.
-                let files = request.crypt_map.as_ref().unwrap();
+                let files = request.cache.crypt_map.as_ref().unwrap();
 
                 if let Ok(info) = FileInfo::from_fd(fd) {
                     let files = files.0.lock().unwrap_or_else(|err| err.into_inner());
@@ -382,7 +383,7 @@ fn syscall_stat_handler(
 
         // We use MUST_PATH, dir refers to the file.
         assert!(
-            path.base.is_empty(),
+            path.base().is_empty(),
             "BUG: MUST_PATH returned a directory for stat, report a bug!"
         );
         let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
@@ -418,7 +419,7 @@ fn syscall_stat_handler(
             }
 
             // Note, unlike statfs, stat does not EINTR.
-            let mut statx = statx(fd, "", flags, mask)?;
+            let mut statx = statx(fd, c"", flags, mask)?;
 
             // SAFETY: Check if the file is a sidechannel device and
             // update its access and modification times to match the
@@ -447,6 +448,11 @@ fn syscall_stat_handler(
                 }
             }
 
+            // If magic path, mask values for easy id.
+            if magic {
+                magic_statx(&mut statx, caps, sflags);
+            }
+
             // SAFETY: The following block creates an immutable byte
             // slice representing the memory of `statx`. We ensure that
             // the slice covers the entire memory of `statx` using
@@ -471,7 +477,7 @@ fn syscall_stat_handler(
 
             // SAFETY: In libc we trust.
             // Note, unlike statfs, stat does not EINTR.
-            let mut stat = fstatat64(Some(fd.as_raw_fd()), "", flags)?;
+            let mut stat = fstatat64(fd, c"", flags)?;
 
             // SAFETY: Check if the file is a sidechannel device and
             // update its access and modification times to match the
@@ -489,6 +495,11 @@ fn syscall_stat_handler(
                 }
             }
 
+            // If magic path, mask values for easy id.
+            if magic {
+                magic_stat(&mut stat, caps, sflags);
+            }
+
             let addr = req.data.args[arg_stat];
             if addr != 0 {
                 if is32 {
@@ -544,3 +555,82 @@ fn syscall_stat_handler(
         Ok(request.return_syscall(0))
     })
 }
+
+fn magic_stat(stat: &mut FileStat64, caps: Capability, flags: Flags) {
+    stat.st_ino = 0;
+    stat.st_nlink = caps.nlink().into();
+    stat.st_mode = magic_mode(caps, flags).into();
+    stat.st_rdev = API_VERSION.dev();
+    stat.st_atime = 505958400; // must match libsyd!
+    stat.st_ctime = -2036448000; // ditto!
+    stat.st_mtime = -842745600; // ditto!
+}
+
+fn magic_statx(statx: &mut FileStatx, caps: Capability, flags: Flags) {
+    statx.stx_ino = 0;
+    statx.stx_nlink = caps.nlink();
+    statx.stx_mode = magic_mode(caps, flags);
+    statx.stx_rdev_major = API_VERSION.major().into();
+    statx.stx_rdev_minor = API_VERSION.minor().into();
+    statx.stx_atime = FileStatxTimestamp {
+        tv_sec: 505958400, // must match libsyd!
+        ..Default::default()
+    };
+    statx.stx_ctime = FileStatxTimestamp {
+        tv_sec: -2036448000, // ditto!
+        ..Default::default()
+    };
+    statx.stx_mtime = FileStatxTimestamp {
+        tv_sec: -842745600, // ditto!
+        ..Default::default()
+    };
+}
+
+#[allow(clippy::cognitive_complexity)]
+#[allow(clippy::cast_possible_truncation)]
+fn magic_mode(caps: Capability, flags: Flags) -> u16 {
+    // Start with file type = character device
+    let mut mode: u16 = libc::S_IFCHR as u16;
+
+    // Special bits
+    if flags.contains(Flags::FL_UNSHARE_MOUNT) {
+        mode |= libc::S_ISVTX as u16;
+    }
+    if flags.contains(Flags::FL_UNSHARE_USER) {
+        mode |= libc::S_ISUID as u16;
+    }
+    if flags.contains(Flags::FL_UNSHARE_NET) {
+        mode |= libc::S_ISGID as u16;
+    }
+
+    // Permission bits for owner/group/other
+    if caps.contains(Capability::CAP_READ) {
+        mode |= libc::S_IRUSR as u16;
+    }
+    if caps.contains(Capability::CAP_WRITE) {
+        mode |= libc::S_IWUSR as u16;
+    }
+    if caps.contains(Capability::CAP_EXEC) {
+        mode |= libc::S_IXUSR as u16;
+    }
+    if caps.contains(Capability::CAP_STAT) {
+        mode |= libc::S_IRGRP as u16;
+    }
+    if caps.contains(Capability::CAP_PROXY) {
+        mode |= libc::S_IWGRP as u16;
+    }
+    if caps.contains(Capability::CAP_TPE) {
+        mode |= libc::S_IXGRP as u16;
+    }
+    if caps.contains(Capability::CAP_LOCK) {
+        mode |= libc::S_IROTH as u16;
+    }
+    if caps.contains(Capability::CAP_CRYPT) {
+        mode |= libc::S_IWOTH as u16;
+    }
+    if caps.contains(Capability::CAP_FORCE) {
+        mode |= libc::S_IXOTH as u16;
+    }
+
+    mode
+}
