@@ -13,7 +13,7 @@ use std::{
 };
 
 use libseccomp::ScmpNotifResp;
-use nix::{errno::Errno, fcntl::AtFlags, unistd::fchdir, NixPath};
+use nix::{errno::Errno, fcntl::AtFlags, NixPath};
 
 use crate::{
     compat::{
@@ -21,6 +21,7 @@ use crate::{
         XATTR_SIZE_MAX,
     },
     config::ROOT_FILE,
+    cookie::safe_fchdir,
     kernel::{syscall_path_handler, to_atflags},
     lookup::FsFlags,
     path::XPath,
@@ -100,9 +101,8 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         // and allocating a Vector of that capacity may overflow.
         let len = to_len_cap(req.data.args[3], XATTR_SIZE_MAX)?;
         let mut buf = if len > 0 {
-            let mut buf = Vec::new();
+            let mut buf: Vec<u8> = Vec::new();
             buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-            buf.resize(len, 0);
             Some(buf)
         } else {
             None
@@ -114,7 +114,7 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
 
         #[expect(clippy::cast_sign_loss)]
         // SAFETY: In libc we trust.
-        let n = match Errno::result(unsafe {
+        let mut n = match Errno::result(unsafe {
             libc::fgetxattr(fd.as_raw_fd(), name, ptr.cast(), len)
         }) {
             Ok(n) => n as usize,
@@ -127,8 +127,12 @@ pub(crate) fn sys_fgetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             Err(errno) => return Err(errno),
         };
 
-        if let Some(buf) = buf {
-            request.write_mem(&buf[..n], req.data.args[2])?;
+        if let Some(mut buf) = buf {
+            // SAFETY: n was returned by the host Linux kernel.
+            unsafe { buf.set_len(n) };
+            buf.shrink_to_fit();
+
+            n = request.write_mem(&buf, req.data.args[2])?;
         }
 
         #[expect(clippy::cast_possible_wrap)]
@@ -217,10 +221,10 @@ pub(crate) fn sys_lsetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         #[expect(clippy::disallowed_methods)]
         let path = path_args.0.as_ref().unwrap();
 
-        let base = if path.base.is_empty() {
+        let base = if path.base().is_empty() {
             XPath::from_bytes(b".")
         } else {
-            path.base
+            path.base()
         };
 
         let req = request.scmpreq;
@@ -250,12 +254,9 @@ pub(crate) fn sys_lsetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             (None, 0)
         } else {
             let len = to_len_val(req.data.args[3], XATTR_SIZE_MAX)?;
-            let mut buf = Vec::new();
-            buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-            buf.resize(len, 0);
-            let n = request.read_mem(&mut buf, req.data.args[2])?;
-            buf.truncate(n);
-            (Some(buf), n)
+            let buf = request.read_vec(req.data.args[2], len)?;
+            let len = buf.len();
+            (Some(buf), len)
         };
         let buf = buf.as_ref().map_or(std::ptr::null(), |b| b.as_ptr()) as *const libc::c_void;
 
@@ -265,11 +266,11 @@ pub(crate) fn sys_lsetxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
         match &path.dir {
             Some(fd) => {
                 // SAFETY: We use fchdir which is TOCTOU-free!
-                fchdir(fd)?;
+                safe_fchdir(fd)?;
             }
             None => {
                 // SAFETY: `/` is never a symlink!
-                fchdir(ROOT_FILE())?;
+                safe_fchdir(ROOT_FILE())?;
             }
         };
 
@@ -351,7 +352,6 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             let mut buf = if len > 0 {
                 let mut buf = Vec::new();
                 buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-                buf.resize(len, 0);
                 Some(buf)
             } else {
                 None
@@ -363,7 +363,7 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
 
             #[expect(clippy::cast_sign_loss)]
             // SAFETY: In libc we trust.
-            let n = match Errno::result(unsafe { libc::flistxattr(fd.as_raw_fd(), ptr, len) }) {
+            let mut n = match Errno::result(unsafe { libc::flistxattr(fd.as_raw_fd(), ptr, len) }) {
                 Ok(n) => n as usize,
                 Err(Errno::ERANGE) if len == XATTR_LIST_MAX => {
                     // SAFETY: Avoid a well-behaving process from
@@ -374,20 +374,22 @@ pub(crate) fn sys_flistxattr(request: UNotifyEventRequest) -> ScmpNotifResp {
                 Err(errno) => return Err(errno),
             };
 
-            let n = if let Some(buf) = buf {
-                // SAFETY: Filter out attributes that start with "user.syd".
-                // SAFETY: Deny only if the Sandbox is locked for the process.
+            if let Some(mut buf) = buf {
+                // SAFETY: n was returned by the host Linux kernel.
+                unsafe { buf.set_len(n) };
+                buf.shrink_to_fit();
+
+                // SAFETY:
+                // Filter out attributes that start with "user.syd".
+                // Deny only if the Sandbox is locked for the process.
                 let buf = if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
-                    Cow::Owned(filterxattr(&buf[..n], n)?)
+                    Cow::Owned(filterxattr(&buf, n)?)
                 } else {
-                    Cow::Borrowed(&buf[..n])
+                    Cow::Borrowed(&buf)
                 };
 
-                request.write_mem(&buf, req.data.args[1])?;
-                buf.len()
-            } else {
-                n
-            };
+                n = request.write_mem(&buf, req.data.args[1])?;
+            }
 
             #[expect(clippy::cast_possible_wrap)]
             Ok(request.return_syscall(n as i64))
@@ -522,10 +524,10 @@ pub(crate) fn sys_lremovexattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             #[expect(clippy::disallowed_methods)]
             let path = path_args.0.as_ref().unwrap();
 
-            let base = if path.base.is_empty() {
+            let base = if path.base().is_empty() {
                 XPath::from_bytes(b".")
             } else {
-                path.base
+                path.base()
             };
 
             let req = request.scmpreq;
@@ -551,11 +553,11 @@ pub(crate) fn sys_lremovexattr(request: UNotifyEventRequest) -> ScmpNotifResp {
             match &path.dir {
                 Some(fd) => {
                     // SAFETY: We use fchdir which is TOCTOU-free!
-                    fchdir(fd)?
+                    safe_fchdir(fd)?
                 }
                 None => {
                     // SAFETY: `/` is never a symlink!
-                    fchdir(ROOT_FILE())?;
+                    safe_fchdir(ROOT_FILE())?;
                 }
             };
 
@@ -621,10 +623,10 @@ fn syscall_getxattr_handler(
     #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
-    let base = if path.base.is_empty() {
+    let base = if path.base().is_empty() {
         XPath::from_bytes(b".")
     } else {
-        path.base
+        path.base()
     };
 
     let name = if req.data.args[1] != 0 {
@@ -650,9 +652,8 @@ fn syscall_getxattr_handler(
     // and allocating a Vector of that capacity may overflow.
     let len = to_len_cap(req.data.args[3], XATTR_SIZE_MAX)?;
     let mut buf = if len > 0 {
-        let mut buf = Vec::new();
+        let mut buf: Vec<u8> = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
         Some(buf)
     } else {
         None
@@ -666,9 +667,9 @@ fn syscall_getxattr_handler(
     match &path.dir {
         Some(fd) => {
             // SAFETY: We use fchdir which is TOCTOU-free!
-            fchdir(fd)?;
+            safe_fchdir(fd)?;
         }
-        None => fchdir(ROOT_FILE())?,
+        None => safe_fchdir(ROOT_FILE())?,
     };
 
     let res = base
@@ -676,7 +677,7 @@ fn syscall_getxattr_handler(
         .with_nix_path(|cstr| unsafe { libc::lgetxattr(cstr.as_ptr(), name, ptr.cast(), len) })?;
 
     #[expect(clippy::cast_sign_loss)]
-    let n = match Errno::result(res) {
+    let mut n = match Errno::result(res) {
         Ok(n) => n as usize,
         Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
             // SAFETY: Avoid a well-behaving process from
@@ -687,8 +688,12 @@ fn syscall_getxattr_handler(
         Err(errno) => return Err(errno),
     };
 
-    if let Some(buf) = buf {
-        request.write_mem(&buf[..n], req.data.args[2])?;
+    if let Some(mut buf) = buf {
+        // SAFETY: n was returned by the host Linux kernel.
+        unsafe { buf.set_len(n) };
+        buf.shrink_to_fit();
+
+        n = request.write_mem(&buf, req.data.args[2])?;
     }
 
     #[expect(clippy::cast_possible_wrap)]
@@ -707,14 +712,14 @@ fn syscall_getxattrat_handler(
     #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
-    let base = if path.base.is_empty() {
+    let base = if path.base().is_empty() {
         XPath::from_bytes(b".")
     } else {
-        path.base
+        path.base()
     };
 
     // Read struct xattr_args which holds the return pointer, buffer size and flags.
-    let mut args = MaybeUninit::<XattrArgs>::uninit();
+    let mut args = MaybeUninit::<XattrArgs>::zeroed();
 
     // SAFETY: Ensure size of XattrArgs matches with user argument.
     if req.data.args[5] != size_of::<XattrArgs>() as u64 {
@@ -729,7 +734,7 @@ fn syscall_getxattrat_handler(
     };
 
     // Read the remote data structure.
-    request.read_mem(buf, req.data.args[4])?;
+    request.read_mem(buf, req.data.args[4], buf.len())?;
 
     // SAFETY: read_mem() has initialized `args` if it succeeded.
     let args = unsafe { args.assume_init() };
@@ -764,7 +769,6 @@ fn syscall_getxattrat_handler(
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
         Some(buf)
     } else {
         None
@@ -788,20 +792,24 @@ fn syscall_getxattrat_handler(
     // SAFETY:
     // 1. `name` is a valid raw pointer (may be NULL)!
     // 2. We do not resolve symbolic links here!
-    let n = match unsafe { getxattrat(fd, base, name, &mut my_args, AtFlags::AT_SYMLINK_NOFOLLOW) }
-    {
-        Ok(n) => n,
-        Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
-            // SAFETY: Avoid a well-behaving process from
-            // repeating calls to potentially exhaust memory.
-            // See tar's tests for an example.
-            return Err(Errno::E2BIG);
-        }
-        Err(errno) => return Err(errno),
-    };
+    let mut n =
+        match unsafe { getxattrat(fd, base, name, &mut my_args, AtFlags::AT_SYMLINK_NOFOLLOW) } {
+            Ok(n) => n,
+            Err(Errno::ERANGE) if len == XATTR_SIZE_MAX => {
+                // SAFETY: Avoid a well-behaving process from
+                // repeating calls to potentially exhaust memory.
+                // See tar's tests for an example.
+                return Err(Errno::E2BIG);
+            }
+            Err(errno) => return Err(errno),
+        };
 
-    if let Some(buf) = buf {
-        request.write_mem(&buf[..n], args.value)?;
+    if let Some(mut buf) = buf {
+        // SAFETY: n was returned by the host Linux kernel.
+        unsafe { buf.set_len(n) };
+        buf.shrink_to_fit();
+
+        n = request.write_mem(&buf, args.value)?;
     }
 
     #[expect(clippy::cast_possible_wrap)]
@@ -846,12 +854,9 @@ fn syscall_setxattr_handler(
         (None, 0)
     } else {
         let len = to_len_val(req.data.args[3], XATTR_SIZE_MAX)?;
-        let mut buf = Vec::new();
-        buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
-        let n = request.read_mem(&mut buf, req.data.args[2])?;
-        buf.truncate(n);
-        (Some(buf), n)
+        let buf = request.read_vec(req.data.args[2], len)?;
+        let len = buf.len();
+        (Some(buf), len)
     };
     let buf = buf.as_ref().map_or(std::ptr::null(), |b| b.as_ptr()) as *const libc::c_void;
 
@@ -875,14 +880,14 @@ fn syscall_setxattrat_handler(
     #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
-    let base = if path.base.is_empty() {
+    let base = if path.base().is_empty() {
         XPath::from_bytes(b".")
     } else {
-        path.base
+        path.base()
     };
 
     // Read struct xattr_args which holds the extension name, buffer size and flags.
-    let mut args = MaybeUninit::<XattrArgs>::uninit();
+    let mut args = MaybeUninit::<XattrArgs>::zeroed();
 
     // SAFETY: Ensure size of XattrArgs matches with user argument.
     if req.data.args[5] != size_of::<XattrArgs>() as u64 {
@@ -897,7 +902,7 @@ fn syscall_setxattrat_handler(
     };
 
     // Read the remote data structure.
-    request.read_mem(buf, req.data.args[4])?;
+    request.read_mem(buf, req.data.args[4], buf.len())?;
 
     // SAFETY: read_mem() has initialized `args` if it succeeded.
     let args = unsafe { args.assume_init() };
@@ -927,12 +932,9 @@ fn syscall_setxattrat_handler(
         (None, 0)
     } else {
         let len = to_len_val(args.size.into(), XATTR_SIZE_MAX)?;
-        let mut buf = Vec::new();
-        buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
-        let n = request.read_mem(&mut buf, args.value)?;
-        buf.truncate(n);
-        (Some(buf), n)
+        let buf = request.read_vec(args.value, len)?;
+        let len = buf.len();
+        (Some(buf), len)
     };
     let buf = buf.as_ref().map_or(std::ptr::null(), |b| b.as_ptr()) as *const libc::c_void;
 
@@ -967,10 +969,10 @@ fn syscall_listxattr_handler(
     #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
-    let base = if path.base.is_empty() {
+    let base = if path.base().is_empty() {
         XPath::from_bytes(b".")
     } else {
-        path.base
+        path.base()
     };
 
     // SAFETY: The size argument to the llistxattr call
@@ -980,7 +982,6 @@ fn syscall_listxattr_handler(
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
         Some(buf)
     } else {
         None
@@ -993,9 +994,9 @@ fn syscall_listxattr_handler(
     match &path.dir {
         Some(fd) => {
             // SAFETY: We use fchdir which is TOCTOU-free!
-            fchdir(fd)?;
+            safe_fchdir(fd)?;
         }
-        None => fchdir(ROOT_FILE())?,
+        None => safe_fchdir(ROOT_FILE())?,
     };
 
     let res = base
@@ -1014,17 +1015,21 @@ fn syscall_listxattr_handler(
         Err(errno) => return Err(errno),
     };
 
-    if let Some(buf) = buf {
-        // SAFETY: Filter out attributes that start with "user.syd".
-        // SAFETY: Deny only if the Sandbox is locked for the process.
+    if let Some(mut buf) = buf {
+        // SAFETY: n was returned by the host Linux kernel.
+        unsafe { buf.set_len(n) };
+        buf.shrink_to_fit();
+
+        // SAFETY:
+        // Filter out attributes that start with "user.syd".
+        // Deny only if the Sandbox is locked for the process.
         let buf = if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
-            Cow::Owned(filterxattr(&buf[..n], n)?)
+            Cow::Owned(filterxattr(&buf, n)?)
         } else {
-            Cow::Borrowed(&buf[..n])
+            Cow::Borrowed(&buf)
         };
 
-        request.write_mem(&buf, req.data.args[1])?;
-        n = buf.len();
+        n = request.write_mem(&buf, req.data.args[1])?;
     }
 
     #[expect(clippy::cast_possible_wrap)]
@@ -1043,10 +1048,10 @@ fn syscall_listxattrat_handler(
     #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
-    let base = if path.base.is_empty() {
+    let base = if path.base().is_empty() {
         XPath::from_bytes(b".")
     } else {
-        path.base
+        path.base()
     };
 
     // SAFETY: The size argument to the llistxattr call
@@ -1056,7 +1061,6 @@ fn syscall_listxattrat_handler(
     let mut buf = if len > 0 {
         let mut buf = Vec::new();
         buf.try_reserve(len).or(Err(Errno::ENOMEM))?;
-        buf.resize(len, 0);
         Some(buf)
     } else {
         None
@@ -1085,17 +1089,21 @@ fn syscall_listxattrat_handler(
         Err(errno) => return Err(errno),
     };
 
-    if let Some(buf) = buf {
-        // SAFETY: Filter out attributes that start with "user.syd".
-        // SAFETY: Deny only if the Sandbox is locked for the process.
+    if let Some(mut buf) = buf {
+        // SAFETY: n was returned by the host Linux kernel.
+        unsafe { buf.set_len(n) };
+        buf.shrink_to_fit();
+
+        // SAFETY:
+        // Filter out attributes that start with "user.syd".
+        // Deny only if the Sandbox is locked for the process.
         let buf = if !sandbox.flags.allow_unsafe_xattr() && sandbox.locked_for(req.pid()) {
-            Cow::Owned(filterxattr(&buf[..n], n)?)
+            Cow::Owned(filterxattr(&buf, n)?)
         } else {
-            Cow::Borrowed(&buf[..n])
+            Cow::Borrowed(&buf)
         };
 
-        request.write_mem(&buf, req.data.args[3])?;
-        n = buf.len();
+        n = request.write_mem(&buf, req.data.args[3])?;
     }
 
     #[expect(clippy::cast_possible_wrap)]
@@ -1151,10 +1159,10 @@ fn syscall_removexattrat_handler(
     #[expect(clippy::disallowed_methods)]
     let path = args.0.as_ref().unwrap();
 
-    let base = if path.base.is_empty() {
+    let base = if path.base().is_empty() {
         XPath::from_bytes(b".")
     } else {
-        path.base
+        path.base()
     };
 
     let name = if req.data.args[3] != 0 {
